// gsn_asymp.hpp

#ifndef GSNASYMP_HPP
#define GSNASYMP_HPP

#include "specialfunc.hpp"

Complex gsn_asymptotic_derivative_initial_sum_term_1(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_2(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_3(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_4(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_5(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_6(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_7(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_8(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_9(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_10(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_11(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);
Complex gsn_asymptotic_derivative_initial_sum_term_12(const double &a, const int &s, const int &mTemp, const double &omega, const double &lambda, const double &r);

#endif
